/*
 * searcher.h
 *
 *  Created on: 2010/09/16
 *      Author: shu
 */

#ifndef SEARCHER_H_
#define SEARCHER_H_

#include "aligner.h"
#include "encoder.h"
#include "sequence_type.h"
#include "score_matrix.h"
#include "sequence_index.h"
#include "db.h"
#include "gapped_extender.h"
#include <stdint.h>
#include <vector>
#include <stack>

using namespace std;

class AlignerParameters;
class Query;

class Searcher {
public:
  Searcher(const AlignerParameters &parameters);
  virtual ~Searcher();
  void Set(Query &query, Db &db);
  void Search(vector<Hit> &hits);

private:
  typedef struct {
    uint32_t query_start;
    uint32_t query_end;
    uint32_t db_start;
    uint32_t db_end;
    uint32_t offset;
    int score;
    int max_score;
  } Node;

  const static uint32_t kMinSearchLength = 3;

  void InitNodeStack();

  void SetNextNodes(const Node &node);

  bool Filter(uint32_t queries_offset, uint32_t db_offset);

  uint8_t max_regular_letter_;
  uint32_t search_length_;
  int threshold_;
  int cutoff_;
  uint32_t max_number_alignment_;
  ScoreMatrix score_matrix_;
  vector<int> matrix_;
  SequenceIndex *query_index_;
  SequenceIndex  *db_index_;
  stack<Node> node_stack_;
  vector<int> score_cutoff_list_;
  uint32_t query_length_;
  uint32_t db_length_;
  Code *query_seq_;
  Code *db_seq_;
  int ungapped_extend_nominal_cutoff_;
  int gapped_extend_nominal_trigger_;
  vector<uint32_t> query_start_list_;
  vector<uint32_t> query_end_list_;
  vector<uint32_t> db_start_list_;
  vector<uint32_t> db_end_list_;

};

#endif /* SEARCHER_H_ */
